/*
 * $Id: XercesVerifierFactory.java,v 1.5 2003/05/31 00:32:49 Bear Exp $
 *
 * Copyright 2002 Kohsuke Kawaguchi, Inc. All Rights Reserved.
 * 
 * Use is subject to license terms.
 */
package org.kohsuke.jarv.xerces;

import java.io.IOException;
import java.util.Vector;

import org.apache.xerces.impl.XMLErrorReporter;
import org.apache.xerces.impl.xs.SchemaGrammar;
import org.apache.xerces.impl.xs.XSDDescription;
import org.apache.xerces.impl.xs.XSGrammarBucket;
import org.apache.xerces.impl.xs.traversers.XSDHandler;
import org.apache.xerces.util.DefaultErrorHandler;
import org.apache.xerces.util.SymbolTable;
import org.apache.xerces.util.XMLGrammarPoolImpl;
import org.apache.xerces.xni.XMLResourceIdentifier;
import org.apache.xerces.xni.XNIException;
import org.apache.xerces.xni.parser.XMLComponentManager;
import org.apache.xerces.xni.parser.XMLEntityResolver;
import org.apache.xerces.xni.parser.XMLErrorHandler;
import org.apache.xerces.xni.parser.XMLInputSource;
import org.iso_relax.verifier.Schema;
import org.iso_relax.verifier.VerifierConfigurationException;
import org.iso_relax.verifier.VerifierFactory;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * VerifierFactory implementation for Xerces-2.
 * 
 * Features and properties set through the VerifierFactory
 * interfaces are passed to the Xerces-2 components.
 * For example, you can write:
 * <pre>
 * factory.setFeature("http://apache.org/xml/features/validation/schema-full-checking",true);
 * </pre>
 * to turn on the full schema checking.
 * 
 * @author Kohsuke KAWAGUCHI (kk@kohsuke.org)
 */
public class XercesVerifierFactory extends VerifierFactory 
    implements XMLEntityResolver {
    
    /** used to parse schemas. */
    private final SymbolTable symbolTable = new SymbolTable();
    
    // XSDHandler needs various helper classes.
    private final XMLErrorReporter errorReporter = new XMLErrorReporter();
    // TODO: implement a draconian error handler
    private final XMLErrorHandler errorHandler = new DefaultErrorHandler();
    
    public XercesVerifierFactory() {
        // initialize helper components
        XMLComponentManager manager = new XMLComponentManagerImpl();
        errorReporter.reset(manager);
    }
    public Schema compileSchema(InputSource is)
        throws VerifierConfigurationException, SAXException, IOException {
            
        // TODO: is this object reusable? is this thread-safe?
        // why am I passing XSGrammarBucket? How is it used?
        XSDHandler handler = new XSDHandler(new XSGrammarBucket());
        
        // initialize the schema compiler.
        // TODO: what about other parameters?
        handler.reset(errorReporter,this,symbolTable,
            new XMLGrammarPoolImpl(),true,false);
        
        // prepare XSDDescription.
        XSDDescription xdecl = new XSDDescription();
        xdecl.setContextType( XSDDescription.CONTEXT_PREPARSE );
        
        String sid = is.getSystemId();
        if(sid!=null) {
            xdecl.setLiteralSystemId(sid);
            xdecl.setExpandedSystemId(sid);
            xdecl.setLocationHints( new String[]{sid} );
        }
         
        // TODO: entity resolution? error handling?
        // probably through the reset method
        SchemaGrammar schema = 
            handler.parseSchema(createXMLInputSource(is),xdecl,null);
        if(schema==null)
            // compile error
            throw new VerifierConfigurationException("unable to compile the schema");
        
        // quick hack. if the schema imports no other schemas,
        if(schema.getImportedGrammars()==null)
            schema.setImportedGrammars(new Vector());
        
        return new XercesSchema(new SchemaGrammar[]{schema});
    }

    /** constructs an XMLInputSource from SAX InputSource. */
    private static XMLInputSource createXMLInputSource( InputSource is ) {
//        with my patch to Xerces-2
//        return new XMLInputSource(is);

//        but for now,
        XMLInputSource xis = new XMLInputSource(
            is.getPublicId(),is.getSystemId(),null);
        xis.setByteStream(is.getByteStream());
        xis.setCharacterStream(is.getCharacterStream());
        xis.setEncoding(is.getEncoding());
        return xis;
    }
    
    private class XMLComponentManagerImpl implements XMLComponentManager
    {
        public Object getProperty( String propName ) {
        // TODO: XMLEntityResolver ENTITY_MANAGER for what?
            if( propName.equals(XercesConstants.ERROR_REPORTER) )
                return errorReporter;
            if( propName.equals(XercesConstants.SYMBOL_TABLE) )
                return symbolTable;
            if( propName.equals(XercesConstants.ERROR_HANDLER) )
                return errorHandler;
                
            return null;    // none was found
        }
        public void setProperty( String propName, Object value ) {
            // setting properties are not allowed
            throw new UnsupportedOperationException();
        }
        public boolean getFeature( String propName ) {
            try {
                return XercesVerifierFactory.this.isFeature(propName);
            } catch( SAXException e ) {
                // if the property is unspecified, assume false
                // as the default value
                return false;
            }
        }
    }

//
//
//    XMLEntityResolver implementation
//
//  
    /** resolve entity by using SAX EntityResolver. */
    public XMLInputSource resolveEntity(XMLResourceIdentifier r) throws XNIException,IOException {
        try {
            EntityResolver er = super.getEntityResolver();
            if(er==null)    return null;
            
            InputSource is = er.resolveEntity(r.getPublicId(),r.getExpandedSystemId());
            if(is==null)    return null;
            else            return createXMLInputSource(is);
        } catch(SAXException e) {
            if(e.getException()!=null)
                throw new XNIException(e.getMessage(),e.getException());
            else
                throw new XNIException(e);
        }
    }

}
